/*
 * line control, connection, etc..
 */

#include <config.h>

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/select.h>
#include <fcntl.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sched.h>
#include <signal.h>
#include <errno.h>
#include <pthread.h>
#include <pty.h>
#include <termios.h>
#include <assert.h>

#include <vector>
#include <algorithm>
#include <queue>
#include <map>

#include "thread++.h"

#include "line.h"
#include "modem.h"
#include "packet.h"
#include "util.h"
#include "protocol.h"
#include "global.h"
#include "compress.h"

class line_control line_control;

recursive_mutex line_control::cons_lock;
bool line_control::cons_changed;
line_control::convec line_control::cons;
pthread_t line_control::transmitter_tid, line_control::receiver_tid;

//******************************************************************************************
// plug, connection...

// guess sequence number from partial data...
// works only as long as WINDOW ~< SEQ_MASK/2
static int SEQ_GUESS (int hint, sequence mod)
{
  sequence seq = (mod & ~SEQ_MASK) | hint;
  
  if (mod - seq > 3)
    seq += SEQ_MASK + 1;
  
  return seq;
}

static void update_ack (packet &p, sequence ack)
{
  p.time.stamp ();
  p[0] = MSG_HEAD ((p[0] & MSG_MASK), ack, MSG_SEQ (p[0]));
}

itcp_connection::itcp_connection (const char *src, u8 sport, const char *dst, u8 dport, plug &p)
  : pl(p)
{
  strcpy (src_msn, src); src_port = sport;
  strcpy (dst_msn, dst); dst_port = dport;
  
  comp = 0;
  decomp = 0;
  
  wire_init ();
}

void itcp_connection::dump ()
{
  static const char *states[] = { "S_CLOSED", "S_LISTEN", "S_SEND_SYN", "S_ESTABLISHED" };
  
  printf ("from %s to %s /%d, state %s (type %s), flags %02x\n", src_msn, dst_msn, dst_port,
          states[state], states[init_state], pflags);
  printf ("		send sequence #%d, receive sequence #%d, framesize %d\n",
          seq_recv, seq_send, frame_tx);
  printf ("		%d bytes in, %d bytes out, %d packets in, %d packets out\n",
          bytes_in, bytes_out, packets_in, packets_out);
  
  if (pflags & PFLAG_GZIP)
    {
      printf ("		gzip ");
      if (comp)
        printf ("compressed %lu ", comp->total ());
      else
        printf ("(NOCOMP) ");
      if (decomp)
        printf ("decompressed %lu ", decomp->total ());
      else
        printf ("(NODECOMP) ");
      printf ("\n");
    }
}

void itcp_connection::connect ()
{
  state = S_SEND_SYN;
  init_state = S_LISTEN;
  line_control.add_connection (this);
}

void itcp_connection::listen ()
{
  state = S_LISTEN;
  init_state = S_LISTEN;
  line_control.add_connection (this);
}

// send fin instead &c.
void itcp_connection::stop ()
{
//  state = S_CLOSING; dream on...
  if (state != S_CLOSED)
    {
      line_control.del_connection (this);
      wire_init ();
    }
}

void itcp_connection::reset ()
{
  wire_init ();
  slog (L_INFO, "connection terminated\n");
}

itcp_connection::~itcp_connection ()
{
  pl.disconnect ();
}

void itcp_connection::push_ctrl (u8 ctl)
{
  ctrl.push (ctl);
}

void itcp_connection::wire_init ()
{
  pl.disconnect ();
  
  state = init_state;
  mask_tx = 0;
  frame_tx = 0;
  
  //D //FIXME: need to delete queues and packet treees as well!!
  while (!ctrl.empty ()) ctrl.pop ();
  while (!data.empty ()) data.pop ();
  
  p_recv.erase (p_recv.begin (), p_recv.end ());
  p_send.erase (p_send.begin (), p_send.end ());
  retransmit_time.reset ();
  
  if (comp)
    delete comp, comp = 0;
  
  if (decomp)
    delete decomp, decomp = 0;
}

void itcp_connection::wire_ready ()
{
  slog (L_ITCP, "wire ready.\n");
  
  frame_tx = 0;
  seq_recv = 0;  // already received
  seq_send = 1;  // next to send
  seq_ackn = 0;  // already acknowledged
  pflags = 0;
  bytes_in =
  bytes_out =
  packets_in =
  packets_out = 0;
  do_ack = false;
  state = S_ESTABLISHED;
  push_ctrl (CTL_ATOM_PING);
  pl.connect ();
}

packet itcp_connection::send_packet ()
{
  guard g(lock);
  
  if (!mask_tx && state != S_CLOSED && state != S_LISTEN)
    mask_tx = mask_map.dst_mask (dst_msn);
  
  packet q(src_msn, dst_msn, mask_tx);
  
  if (mask_tx)
    frame_tx = packet_length (mask_tx);
  else
    return q;
  
  int togo = frame_tx;
  
  switch (state)
    {
      default:
        abort ();
      
      case S_LISTEN:
        break;
      
      case S_SEND_SYN:
        q.mask = mask_tx;
        q.put (MSG_SYN);
        q.put (PROTOCOL_VERSION);
        q.put (src_port);
        q.put (dst_port);
        q.put (PFLAG_WANT);
        wire_ready ();
        break;
        
      case S_ESTABLISHED:
        sequence resend = 0;
        
        for (packet_map::const_iterator p = p_send.begin (); p != p_send.end (); ++p)
          if (p->second.time.age () > retransmit_time ())
            {
              slog (L_DEBUG, "DEBUG: packet %d age %ld > %ld\n", p->first, p->second.time.age (), retransmit_time ());
              resend = p->first;
              retransmit_time.next ();
              break;
            }
        
        if (!resend)
          {
            pl.data_poll (this);
            
            if (comp)
              while (comp->avail (0) < frame_tx && data.size ())
                comp->put (data.front ()), data.pop ();
            
            if (comp)//D
              slog (L_INFO, "DATASIZE %d AVAIL %d\n", data.size (), comp->avail (0));
          }
        
        if (resend)
          {
            tx_at (100);
            
            for (packet_map::iterator p = p_send.begin (); p != p_send.end (); ++p)
              update_ack (p->second, seq_recv);
            
            slog (L_ITCP, "INFO: resending %d\n", resend);
            q = p_send[resend];
          }
        else if (p_send.size () < WINDOW
                 && !(ctrl.empty()
                      && (comp ? !comp->avail (frame_tx) : data.empty())))
          {
            do_ack = false;
            int data_size = comp ? comp->avail (frame_tx) : data.size ();
            
            if (comp)//D
              slog (L_INFO, "XATASIZE %d AVAIL %d = %d\n", data.size (), comp->avail (0), data_size);
            
            if (!ctrl.empty () || data_size < frame_tx - 1)
              {
                slog (L_ITCP, "INFO: sending mixed packet r%d, s%d\n", seq_recv, seq_send);
                // we need a mixed packet
                
                q.put (MSG_HEAD (MSG_MIXED, 0, seq_send)); togo--;
                
                while (togo > 0 && !ctrl.empty ())
                  q.put (ctrl.front ()), ctrl.pop (), togo--;
                
                while (togo > data_size + 1)
                  q.put (CTL_ATOM_NOP), togo--;
                
                q.put (CTL_ATOM_END), togo--;
              }
            else
              {
                slog (L_ITCP, "INFO: sending data packet r%d, s%d\n", seq_recv, seq_send);
                // a data packet suffices
                q.put (MSG_HEAD (MSG_DATA, 0, seq_send)); togo--;
              }
            
            while (togo)
              {
                if (comp)
                  q.put (comp->get ());
                else
                  q.put (data.front ()), data.pop ();
                
                bytes_out++, togo--;
              }
            
            assert (q ());
            
            if (q ())
              {
                update_ack (q, seq_recv);
                p_send[seq_send] = q;
                seq_send++;
              }
          }
        else if (do_ack)
          {
            do_ack = false;
            if (p_send.empty ())
              {
                q.put (MSG_HEAD (MSG_ACK, seq_recv, 0));
                slog (L_WARN, "WARNING: sending ack-only %d (%d)\n", seq_recv, seq_send);
              }
            else
              {
                packet_map::reverse_iterator p = p_send.rbegin ();
                q = p->second;
                update_ack (q, seq_recv);
                
                slog (L_ITCP, "resending packet %d for ack\n", p->first);
              }
          }
        else if (p_send.size ())
          {
            slog (L_ITCP, "nothing to send.... (%d, %d/%d) %d packages unacknowledged\n",
                  seq_recv, seq_send, seq_ackn, p_send.size ());
          }
        
        break;
    }
  
  if (q())
    packets_out++;
  
  return q;
}

void itcp_connection::receive_packet (const packet &p)
{
  int ack, seq;
  u8 head = *p.begin ();
  packet_map::iterator pkt;
  
  guard g(lock);
  
  packets_in++;
  
  if (head == MSG_SYN)
    {
      if (state = S_SEND_SYN)
        {
          state = S_ESTABLISHED;
          wire_ready ();
        }
      else if (state = S_ESTABLISHED)
        {
        }
      else
        {
          slog (L_WARN, "WARNING: spurious packet received in S_LISTEN (from %s)\n", p.src_msn);
          control_connection.send_rst (p);
          return;
        }
      
      if (p.size () > 4)
        {
          pflags = p[4] & PFLAG_CAN;
          slog (L_INFO, "connection from %s, requested flags %02x, supported pflags %02x\n",
                p.src_msn, p[4], pflags);
          push_ctrl (CTL_CMD_FLAGS);
          push_ctrl (pflags);
          
          if (!comp && pflags & PFLAG_GZIP)
            comp = new compressor;
        }
      else
        slog (L_WARN, "WARNING: no pflags in connection setup (old freeisdn at %s)\n", p.src_msn);
      
      tx_now ();
    }
  else if ((state == S_SEND_SYN && head == MSG_HEAD (MSG_ACK, 1, 1)) || state == S_ESTABLISHED)
    {
      if ((head & MSG_MASK) == MSG_ACK)
        {
          ack = SEQ_GUESS (MSG_ACK (head), seq_ackn);
//D          slog (L_WARN, "MSG_ACK (%d) <=> %d\n", ack, seq_ackn);
          
          acknowledge (ack);
        }
      else if ((head & MSG_MASK) == MSG_DATA || (head & MSG_MASK) == MSG_MIXED)
        {
          seq = SEQ_GUESS (MSG_SEQ (head), seq_recv);
          ack = SEQ_GUESS (MSG_ACK (head), seq_ackn);
          
          if (state == S_SEND_SYN)
            {
              if (seq <= 1 && ack <= 1) /* first packet.  */
                {
                  wire_ready ();
                }
              else
                {
                  slog (L_WARN, "WARNING: spurious packet received -> RST\n");
                  control_connection.send_rst (p);
                  goto err_out;
                }
            }
          
          do_ack = true;
          
          acknowledge (ack);
          
          if (seq > seq_recv)
            {
              p_recv[seq] = p;
              
              while ((pkt = p_recv.find (seq_recv + 1)) != p_recv.end ())
                {
                  seq_recv++;
                  slog (L_ITCP, "INFO: received sequence number %d\n", seq_recv);
                  push_packet (pkt->second);
                  p_recv.erase (pkt);
                }
            }
          else
            slog (L_ITCP, "INFO: ignoring extra packet %d\n", seq);
        }
      else if (head != MSG_SYN)
        slog (L_WARN, "WARNING: spurious packet received (EST) from %s\n", p.src_msn);
    }
  else
    {
      slog (L_WARN, "WARNING: spurious packet received from %s\n", p.src_msn);
      control_connection.send_rst (p);
    }
  
err_out: ;
}

bool itcp_connection::match (const packet &p)
{
  // reject ctrl-messages
  if ((p[0] & MSG_MASK) == MSG_SYN && (p[0] != MSG_SYN))
    return false;
  
  return 0 == strcmp (dst_msn, p.src_msn);
}

// receive data packets in-sequence (after flow-control)
void itcp_connection::push_packet (const packet &p)
{
  int togo = p.size () - 1;
  packet::const_iterator data = p.begin ();
  
  if ((*data++ & MSG_MASK) == MSG_MIXED)
    {
      /* parse ctl bytes.  */
      for(;;)
        {
          u8 ctl = *data++; togo--;
          switch (ctl & CTL_MASK)
            {
              case CTL_ATOM:
                switch (ctl)
                  {
                    case CTL_ATOM_NOP:
                      break;
                    
                    case CTL_ATOM_PING:
                      push_ctrl (CTL_ATOM_PONG);
                      break;
                    
                    case CTL_ATOM_PONG:
                      slog (L_JUNK, "received PONG\n");
                      break;
                    
                    case CTL_ATOM_FIN:
                      /* tja... */
                      wire_init ();
                      break;
                    
                    case CTL_ATOM_END:
                      goto eo_ctl;
                    
                    default:
                      slog (L_WARN, "unknown ctl atom %02x received, ignoring\n");
                  }
                break;
              
              case CTL_RESEND:
                slog (L_ITCP, "request to resend %d IGNORED\n", ctl & 7);
                break;
              
              case CTL_CMD:
                switch (ctl)
                  {
                    case CTL_CMD_LACK:
                      slog (L_INFO, "large acknowledge %02x received\n", *data);
                      slog (L_FATAL, "FATAL: funny that I don't implement them yet\n");//D
                      break;
                    
                    case CTL_CMD_FLAGS:
                      if (*data != pflags)
                        {
                          pflags = *data;
                          slog (L_INFO, "setting control flags to %02x\n", pflags);
                          push_ctrl (CTL_CMD_FLAGS);
                          push_ctrl (pflags);
                          
                        }
                      else
                        slog (L_INFO, "already have control flags\n");
                      
                      if (pflags & PFLAG_GZIP)
                        {
                          if (!comp)
                            comp = new compressor;
                          if (!decomp)
                            decomp = new decompressor;
                        }
                      
                      break;
                    
                    default:
                      slog (L_WARN, "unknown ctl cmd received (%02x), skipping data\n", ctl);
                }
                data++, togo--;
                break;
              
              default:
                slog (L_WARN, "unknown ctl byte received (%02x), skipping\n", ctl);
            }
        }
      eo_ctl: ;
    }
  
  if (togo)
    {
      if (decomp)
        {
          int push;
          
          while (togo--)
            {
              decomp->put (*data++);
              
              if ((push = decomp->avail ()))
                {
                  bytes_in += push;
                  pl.data_push (decomp->get_buf (), push);
                }
            }
        }
      else
        {
          bytes_in += togo;
          pl.data_push (data, togo);
        }
    }
}

void itcp_connection::acknowledge (sequence ack)
{
  if (ack > seq_ackn) // hmm.. new packets to acknowledge...
    {
      do_ack = true;
      
      while (ack > seq_ackn)
        {
          seq_ackn++;
          
          packet_map::iterator i = p_send.find (seq_ackn);
          
          if (i == p_send.end ())
            {
              slog (L_DEBUG, "DEBUG: NOT acknowledging future packet %d\n", seq_ackn);
              seq_ackn--;
              return;
            }
          
          slog (L_ITCP, "INFO: acknowledging %d\n", seq_ackn);
          retransmit_time.reset ();
          
          p_send.erase (i);
        }
    }
  else if (ack == seq_ackn) // same ack twice in a row, resend unconditionally
    p_send.find (ack) -> second.time.age (retransmit_time ());
}

bool tty_plug::connect()
{
  if (pid)
    {
      log (L_WARN, "already bound!\n");
      return true;
    }
  
  termios term = {
    IGNPAR,
    0,
    CS8 | CREAD | CLOCAL | CRTSCTS | B300,
    0
  };
  term.c_cc[VMIN]=1;
  
  pid = forkpty (&master, ttyname, &term, 0);
  
  if (pid == 0) /* child */
    {
      execl ("/bin/sh", "sh", "-c", command);
      exit (127);
    }
  else if (pid > 0) /* parent */
    {
      tcsetattr (master, TCSANOW, &term);
      log (L_INFO, "pid = %d, amaster = %d, ttyname = %s, cmd = |%s|\n", pid, master, ttyname, command);
      return true;
    }
  else
    return false;
}

void tty_plug::disconnect ()
{
  if (pid)
    {
      close (master);	/* hopefully this is enough.. */
      kill (SIGHUP, pid);
      pid = 0;
    }
}

void tty_plug::data_poll (itcp_connection *con)
{
  assert (pid);
  fcntl (master, F_SETFL, O_NONBLOCK);
  
  for(;;)
    {
      u8 buffer[40];
      int len = 400 - (con->data.size ());
      
      if (len > (int)sizeof buffer)
        len = sizeof buffer;
      
      len = read (master, buffer, len);
      
      if (len <= 0)
        break;
      
      log (L_JUNK, "polled %d bytes\n", len);
      
      for (int i = 0; i < len; i++)
        con->data.push (buffer[i]);
    }
}

void tty_plug::data_push (packet::const_iterator data, int len)
{
  assert (pid);
  
//  printf ("*************************** data_push with len = %d: ", len);//D
  
  fcntl (master, F_SETFL, 0);
  while (len--)
//    {//D
//      printf ("%02x", (int)*data);//D
    write (master, data, 1), ++data;
//    }//D
//  printf ("\n");//D
}

//******************************************************************************************
// isdn scheduler, transmitter & receiver

// Exxxx -> wo bemerkt				retransmit	reached?
// E023F -> analog				NEVER		N
// E021F -> lokal busy				NEVER		N
// E0201 -> 9377343 (unassigned?)		NEVERNEVER	N
// E021c -> rejected (uh-oh, that's good?)	NEVERNEVER	N
// E0222 -> lokal beide busy			NOW		N
// E0211 -> lokal busy				NOW		N
// E0058 -> lukas anderes isdn-endgert?	NO		?		isdn4linux-fehler!!
// E007F -> lokal				NO		Y
// E0212 -> ohne reject, keiner da.		NO		?
// other -> ??					NO		?
static void write_pkt(tty *line, packet &p) {
  char buff[600];
  char res[60];
  int cause;
  
  {
    char pkt[600], *x = pkt;
    for(packet::const_iterator i=p.begin(); i != p.end(); i++)
      x += sprintf (x, "%02x", *i);
    *x = 0;
    clog (L_PKT, p.src_msn,"SPKT %d:%s\n", p.size(), pkt);
  }
  
  sprintf (buff, "ATD~%s\r", (const char *)p);
  
  log (L_JUNK, "CMD %s\n", buff);
  line->cmd (buff, res, 60);
  log (L_LINE, "line response: %s", res);
  assert (sscanf (res, "%*[^/]/%05x\r", &cause) == 1);
//  log (L_LINE, "cause code: %05x\n", cause);
  cause &= 0xffff;
  
  timeval tv = {0, 100000};
  select (0, 0, 0, 0, &tv);
}

void *line_control::transmitter (void *x)
{
  tty *line = (tty *)x;
  packet pkt;
  timestamp tx_last;
  
  for(;;)
    {
      bool do_pause = true;
      
      cons_lock.lock ();
      
      cons_changed = false;
      convec::iterator conp = cons.begin ();
      
      while (conp != cons.end () && !cons_changed)
        {
          connection *con = *conp;
          cons_lock.unlock ();
          
          guard g(con->lock);
          
          if (con->tx_inhibit.past ())
            {
              pkt = con->send_packet ();
              
              if (pkt ())
                {
                   tx_last.stamp ();
                   write_pkt (line, pkt);
                   do_pause = false;
                }
            }
          
          cons_lock.lock ();
          ++conp;
        }
      
      cons_lock.unlock ();
      
      if (do_pause)
        sleep (1);
    }
  
  delete line;
  
  return 0;
}

void *line_control::receiver (void *file)
{
  FILE *ctrl = (FILE *)file;
  char line[200];
  
  while (fgets (line, sizeof line, ctrl))
    {
      log (L_ISDN, "%s", line);
      if (strlen (line)<18 || strncmp (line+9, "PACKET: ", 8))
        continue;
      
      packet p(line+9);
      
      p.time.stamp ();
      
      {
        char pkt[600], *x = pkt;
        for(packet::const_iterator i=p.begin(); i != p.end(); i++)
          x += sprintf (x, "%02x", *i);
        *x = 0;
        clog (L_PKT, p.dst_msn,"RPKT %d:%s\n", p.size(), pkt);
      }
      
      cons_lock.lock ();
      convec::iterator conp = cons.begin ();
      connection *con = &control_connection;
      
      while (conp != cons.end ())
        if ((*conp)->match (p))
          {
            con = *conp;
            break;
          }
        else
          ++conp;
      
      cons_lock.unlock ();
      con->receive_packet (p);
      pthread_kill (transmitter_tid, SIGCHLD);
    }
  
  return 0;
}

void line_control::add_modemdev (const char *path)
{
  tty *line = new tty(path);
  char res[60];
  
  do {
    line->cmd ("AT&FE0&E0S13=9\r", res, 60);
  } while (!strcmp ("OK\r", res));
  
  run (&transmitter_tid, (thread)&line_control::transmitter, line);
}

void line_control::set_ctrldev (const char *path)
{
  FILE *f = fopen ((char *)path, "r");
  
  assert (f); /*D*/
  run (&receiver_tid, (thread)&line_control::receiver, (void *)f);
}

void line_control::add_connection (connection *con)
{
  guard g(cons_lock);
  cons.push_back (con);
  cons_changed = true;
  
  pthread_kill (transmitter_tid, SIGCHLD);
}

void line_control::del_connection (connection *con)
{
  guard g(cons_lock);
  cons_changed = true;
  
  vector<connection *>::iterator c = cons.begin ();
  
  while (c != cons.end ())
    if (*c == con)
      {
        cons.erase (c);
        break;
      }
    else
      ++c;
}

line_control::~line_control ()
{
  void *res;
  
  log (L_INFO, "joining for termination...\n");
  
  pthread_join (receiver_tid, &res);
  pthread_join (transmitter_tid, &res);
}

void line_control::command_all (connection::message msg, const char *dst_msn)
{
  guard g(cons_lock);
  
  vector<connection *>::iterator c = cons.begin ();
  
  while (c != cons.end ())
    {
      if (!strcmp ((*c)->dst_msn, dst_msn))
        ((*c)->*msg) ();
      ++c;
    }
}

